#include "rtprefs.h"

#include <assert.h>

#include <Types.h>
#include <Resources.h>
#include <LowMem.h>
#include <Fonts.h>
#include <TextUtils.h>
#include <Folders.h>
#include <Errors.h>

#include "ossmac.h"



#if PRAGMA_STRUCT_ALIGN
	#pragma options align=mac68k
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(push, 2)
#elif PRAGMA_STRUCT_PACK
	#pragma pack(2)
#endif

struct RTPreferences {
	short	fontSize;
	Str63	fontName;
	Rect	windowUserSize;
	Rect	windowStandardSize;
};

/*
	The way the window rects work is modeled on MPWs MPSR resource, and
	its a little peculiar.  Essentially, windowStandardSize always
	represents the current size of the window.  If windowUserSize is the
	same, the window is unzoomed.  If its different, the window is zoomed,
	and the user state is in windowUserSize.
	
	It makes more sense from the client side -- after resizing the window
	with SizeWindow (i.e., the user resized the window), call RTPSetWindow-
	UserSize; after zooming with ZoomWindow, call RTPSetWindowStandardSize
	with the new size.
*/

#if PRAGMA_STRUCT_ALIGN
	#pragma options align=reset
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(pop)
#elif PRAGMA_STRUCT_PACK
	#pragma pack()
#endif

const short		kRTPreferencesFileStringID = 128;
const OSType	kPrefsOSType = 'Pref';
const short		kPrefsID = 128;

#define			kDefaultFont "\pMonaco"
const short		kDefaultFontSize = 9;

static RTPreferences **gshPrefs = NULL;


static bool GetPrefsHandle()
{
	if (gshPrefs)
		return true;
	else {
		(void) RTPLoad();
		return gshPrefs != NULL;
	}
}

static void DefaultGameWindowSize(Rect& r)
{
	// This routine sets the "default default" window size, which
	// means that it assumes the display font to be Monaco 9.
	
	const short kTitleBarHeight = 16;
	const short kInset = 7;
	
	r.top = LMGetMBarHeight() + kTitleBarHeight + kInset;
	r.left = kInset;
	r.bottom = qd.screenBits.bounds.bottom - kInset;
	r.right = 512 - kInset;
}


void RTPGetFontName(StringPtr name)
{
	if (name) {
		if (GetPrefsHandle()) {
			HLock((Handle) gshPrefs);
			StringCopy_pp(name, (*gshPrefs)->fontName, 0);
			HUnlock((Handle) gshPrefs);
		}
		else
			StringCopy_pp(name, kDefaultFont, 0);
	}
}

short RTPGetFontID()
{
	short	familyID = 0;
	
	if (GetPrefsHandle()) {
		HLock((Handle) gshPrefs);
		GetFNum((*gshPrefs)->fontName, &familyID);
		HUnlock((Handle) gshPrefs);
	}
	if (familyID == 0)
		GetFNum(kDefaultFont, &familyID);
	
	return (familyID == 0) ? applFont : familyID;
}

short RTPGetFontSize()
{
	if (GetPrefsHandle())
		return (*gshPrefs)->fontSize;
	else
		return kDefaultFontSize;
}

void RTPGetWindowSize(Rect& r)
{
	if (GetPrefsHandle())
		r = (*gshPrefs)->windowStandardSize;
	else
		DefaultGameWindowSize(r);
}

void RTPSetFont(ConstStr255Param name)
{
	assert(StrLength(name) < sizeof((*gshPrefs)->fontName));
	
	if (GetPrefsHandle()) {
		HLock((Handle) gshPrefs);
		StringCopy_pp((*gshPrefs)->fontName, name, sizeof((*gshPrefs)->fontName));
		HUnlock((Handle) gshPrefs);
		RTPSave();
	}
}

void RTPSetFontSize(short size)
{
	if (GetPrefsHandle()) {
		(*gshPrefs)->fontSize = size;
		RTPSave();
	}
}

void RTPSetWindowUserSize(const Rect& r)
{
	if (GetPrefsHandle()) {
		(*gshPrefs)->windowUserSize = r;
		(*gshPrefs)->windowStandardSize = r;
		RTPSave();
	}
}

void RTPSetWindowStandardSize(const Rect& r)
{
	if (GetPrefsHandle()) {
		(*gshPrefs)->windowStandardSize = r;
		RTPSave();
	}
}


// ======================================================================

static short	gsPrefsFileID = kResFileNotOpened;
static FSSpec	gsPrefsSpec = { 0, 0, "\p" };

class RTPInitializer {
	public:
		RTPInitializer()
			{ }
		~RTPInitializer()
			{ if (gsPrefsFileID != kResFileNotOpened) CloseResFile(gsPrefsFileID); }
};
static RTPInitializer	gsRTPInit;


static OSErr RTPOpen()
/*
 * ## Open the preferences file, or create it if it doesn't exist.
 */
{
	StringHandle	hPrefsName;
	FSSpec			prefs;
	OSErr			err = noErr;
	
	hPrefsName = GetString(kRTPreferencesFileStringID);
	if (!hPrefsName) {
		err = ResError();
	}
	else {
		HLock((Handle) hPrefsName);
		StringCopy_pp(prefs.name, *hPrefsName, sizeof(prefs.name));
		ReleaseResource((Handle) hPrefsName);
		
		err = FindFolder(kOnSystemDisk, kPreferencesFolderType, false, &prefs.vRefNum, &prefs.parID);
		if (err == noErr) {
			// Try to open the prefs file with exclusive read-write permission.
			gsPrefsFileID = FSpOpenResFile(&prefs, fsRdWrPerm);
			if (gsPrefsFileID == kResFileNotOpened) {
				// That didn't work, try for read-only permission.
				gsPrefsFileID = FSpOpenResFile(&prefs, fsRdPerm);
				if (gsPrefsFileID == kResFileNotOpened) {
					// Still no dice -- what's wrong?
					err = ResError();
					if (err == fnfErr) {
						// The file doesn't exist; we need to create it.
						FSpCreateResFile(&prefs, gAppCreator, 'TEXT', smSystemScript);
						err = ResError();
						if (err == noErr) {
							gsPrefsFileID = FSpOpenResFile(&prefs, fsRdWrPerm);
							if (gsPrefsFileID == kResFileNotOpened) {
								// Created it, but couldn't open it.  (Odd, but possible.)
								err = ResError();
							}
						}
						else {
							// Couldn't create the prefs file.
						}
					}
					else {
						// The file exists, but couldn't be opened.
					}
				}
				else {
					// Prefs file was opened read-only -- warn that changes won't be saved.
				}
			}
		}
	}
	return err;
}


static void FillDefaultPrefs(RTPreferences **hPrefs)
{
	HLock((Handle) hPrefs);
	(*hPrefs)->fontSize = kDefaultFontSize;
	StringCopy_pp((*hPrefs)->fontName, kDefaultFont, sizeof((*hPrefs)->fontName));
	DefaultGameWindowSize((*hPrefs)->windowUserSize);
	(*hPrefs)->windowStandardSize = (*hPrefs)->windowUserSize;
	HUnlock((Handle) hPrefs);
}

OSErr RTPLoad()
/*
 * ## Load the preferences from the preferences file.  If it
 * ## doesn't exist, create it with the default preferences.
 */
{
	if (!gshPrefs) {
		OSErr			err = noErr;
		
		/* Open the preferences file, creating it if necessary.
		*/
		if ((err = RTPOpen()) != noErr) {
			// Couldn't open the preferences file -- warn that changes won't be saved.
		}
		
		/* Get a handle to a preferences record, creating it if necessary.
		*/
		// Try to get the preferences from resources.
		gshPrefs = (RTPreferences **) GetResource(kPrefsOSType, kPrefsID);
		if (!gshPrefs) {
			// None available (either in the prefs file or the application), so create one.
			gshPrefs = (RTPreferences **) NewHandle(sizeof(RTPreferences));
			if (gshPrefs) {
				FillDefaultPrefs(gshPrefs);
			}
			else {
				// NewHandle failed -- bail out now.
				return MemError();
			}
		}
		else {
			// Make sure the handle is the correct size.
			Size	hSize = GetHandleSize((Handle) gshPrefs);
			
			if (hSize < sizeof(RTPreferences)) {
				DPRINTF(("Preferences handle is too small!"));
				FillDefaultPrefs(gshPrefs);
			}
			else if (hSize > sizeof(RTPreferences)) {
				DPRINTF(("Preferences handle is too big!"));
				SetHandleSize((Handle) gshPrefs, sizeof(RTPreferences));
				if (gsPrefsFileID != kResFileNotOpened &&
					HomeResFile((Handle) gshPrefs) == gsPrefsFileID)
				{
					ChangedResource((Handle) gshPrefs);
				}
			}
			else {
				// Just right!
			}
		}
		
		/* If the prefs didn't come from the preferences file and we've got
		 * a prefs file to write to, write them there now.
		 */
		if (gsPrefsFileID != kResFileNotOpened &&
			HomeResFile((Handle) gshPrefs) != gsPrefsFileID)
		{
			if (HomeResFile((Handle) gshPrefs) != -1)
				DetachResource((Handle) gshPrefs);
			
			// Fill out window rects with defaults.
			HLock((Handle) gshPrefs);
			DefaultGameWindowSize((*gshPrefs)->windowUserSize);
			DefaultGameWindowSize((*gshPrefs)->windowStandardSize);
			HUnlock((Handle) gshPrefs);
			
			AddResource((Handle) gshPrefs, kPrefsOSType, kPrefsID, "\p");
			return ResError();
		}
		else
			return noErr;
	}
	else
		// Already loaded.
		return noErr;
}

OSErr RTPSave()
	// Save the preferences back to the preferences file.
{
	if (gshPrefs && gsPrefsFileID != kResFileNotOpened &&
		HomeResFile((Handle) gshPrefs) == gsPrefsFileID)
	{
		ChangedResource((Handle) gshPrefs);
		return ResError();
	}
	else
		return fnfErr;
}

